#Written by Dr. Hicham Badri @Mobius Labs GmbH - 2023
#####################################################
import torch
import numpy as np 

#re-estimate teh scale based on the inverse median
def update_scale_inverse_median(W_f, scale, zero, axis, min_max):
	scale_rng = 2e4
	z_val     = 1e-4
	delta     = 1e-2

	W_q       = torch.round(W_f*scale + zero).clamp(min_max[0], min_max[1])

	#Correct zero to avoid W_q==zero
	zero_c              = zero.clone()
	zero_c_indx         = torch.sum(1.*((W_q - zero)==0), axis=axis, keepdim=True)>0
	zero_c[zero_c_indx] = zero_c[zero_c_indx] + delta

	#Build scale tensor
	W_f_c               = W_f.clone()
	W_f_c_mask          = torch.abs(W_f_c)<z_val
	W_f_c[W_f_c_mask]   = z_val

	scale_tensor        = (W_q - zero_c).float()/W_f_c.float()
	# W_r               = (W_q - zero_c)/scale_tensor

	#Normalize scale_tensor
	scale_b = torch.median(scale_tensor, axis=axis, keepdim=True)[0]
	scale_b = scale_b.clamp(min=-scale_rng, max=scale_rng).half()

	#Mix with older scale
	W_r     = (W_q - zero_c)/scale_b
	err_b   = torch.abs(W_f - W_r).mean(axis=axis, keepdim=True)

	W_r     = (W_q - zero_c)/scale
	err_a   = torch.abs(W_f - W_r).mean(axis=axis, keepdim=True)

	mask    = (err_b < err_a).half()
	scale_b = mask*scale_b + (1 - mask)*scale

	#W_r   = (W_q - zero_c)/scale_b
	return scale_b, zero_c


#Greedy local search 
def update_scale_grid_search(W_f, scale, zero, axis, min_max):
	N             = 128 + 1 #Make sure it's an odd number so that the original scale is included
	rng_dump      = 0.05 #0.05 / 1.
	z_val         = 2e-4

	device        = scale.device
	dtype         = scale.dtype
	###############################
	W_q           = torch.round(W_f*scale + zero).clamp(min_max[0], min_max[1])
	n_clusters    = max(W_q.shape[0], W_q.shape[1])
	rng           = torch.abs(scale).mean()*rng_dump if (rng_dump<1.) else rng_dump

	scale_shifted = torch.linspace(-rng, rng, N)[:,None].to(dtype).to(device).repeat(1, n_clusters) + scale

	#Safe inverse
	scale_shifted[torch.logical_and(scale_shifted>=0, torch.abs(scale_shifted)<=z_val)] = z_val 
	scale_shifted[torch.logical_and(scale_shifted<0,  torch.abs(scale_shifted)<=z_val)] = -z_val 

	err = torch.empty([N, n_clusters], dtype=dtype, device=device)
	for i in range(N):
		W_r    = (W_q - zero)/scale_shifted[i][None,:]
		err[i] = torch.abs(W_f - W_r).mean(axis=axis, keepdim=True)

	ind_r   = torch.argmin(err, axis=axis).to(torch.int32)
	ind_c   = torch.arange(len(ind_r), dtype=torch.int32, device=device)
	scale_b = scale_shifted[ind_r, ind_c]

	return scale_b


#Proximal solver || W - dequantize(quantize(W))||_p^p - Experimental
@torch.inference_mode()
def optimize_weights_proximal_v2(tensor, scale, zero, min_max, axis=0, device='cuda', 
								opt_params={'lp_norm':0.7, 'beta':1e1, 'kappa':1.01, 'iters':20, 'tol':0., 'early_stop':True, 'scale_gridsearch':False}, 
								verbose=False):
	
	#Params 
	lp_norm    = max(opt_params['lp_norm'], 0.1) 
	beta       = opt_params['beta']
	kappa      = opt_params['kappa']
	iters      = opt_params['iters']
	early_stop = opt_params['early_stop']
	tol        = opt_params['tol']

	#Check
	assert lp_norm<=1., "lp_norm should be <=1"
	assert beta>0.,     "beta should be > 0"
	assert kappa>1.,    "kappa should be > 1"
	assert iters>1,     "iters should be > 1"

	#Cast/device
	dtype      = torch.float16 if (device=='cuda') else torch.float32
	W_f        = tensor.to(dtype).to(device)
	scale      = scale.to(dtype).to(device)
	zero       = zero.to(dtype).to(device)

	if(lp_norm==1):
		shrink_op = lambda x, beta: torch.sign(x)*torch.nn.functional.relu(torch.abs(x) - 1./beta) 
	else:
		shrink_op = lambda x, beta,p=lp_norm: torch.sign(x)*torch.nn.functional.relu(torch.abs(x) - (1./beta)*torch.pow(torch.abs(x), p-1))

	#Update scale: works slightly better. Tested on Llama2 only
	if(opt_params['scale_gridsearch']):
		scale = update_scale_grid_search(W_f, scale, zero, axis, min_max)

	#Optimize for zero-point
	best_error = 1e4
	scale_prev, zero_prev = scale.clone(), zero.clone()
	for i in range(iters):		
		W_q   = torch.round(W_f*scale + zero).clamp(min_max[0], min_max[1])
		W_r   = (W_q - zero)/scale
		
		#current_error = float(torch.pow(torch.abs(W_f - W_r), max(0.80, lp_norm)).mean())
		current_error = float(torch.abs(W_f - W_r).mean())

		if(verbose): 
			print(i, np.round(current_error, 6))

		if(early_stop):
			if(best_error - current_error > tol):
				best_error = current_error
				scale_prev, zero_prev = scale.clone(), zero.clone()
			else:
				scale, zero = scale_prev.clone(), zero_prev.clone()
				break

		W_e   = shrink_op(W_f - W_r, beta)
		zero  = torch.mean(W_q - (W_f - W_e)*scale, axis=axis, keepdim=True)
		beta *= kappa


	#Clean-up
	scale = scale.to(tensor.device)
	zero  = zero.to(tensor.device)
	del W_f, W_q, W_r, W_e, scale_prev, zero_prev
	torch.cuda.empty_cache()

	return scale, zero 

#Proximal solver || W - dequantize(quantize(W))||_p^p
@torch.inference_mode()
def optimize_weights_proximal_legacy(tensor, scale, zero, min_max, axis=0, device='cuda', opt_params={'lp_norm':0.7, 'beta':1e1, 'kappa':1.01, 'iters':20}, verbose=False):
	lp_norm, beta, kappa, iters = opt_params['lp_norm'], opt_params['beta'], opt_params['kappa'], opt_params['iters']

	dtype  = torch.float16 if (device=='cuda') else torch.float32
	W_f    = tensor.to(dtype).to(device)
	scale  = scale.to(dtype).to(device)
	zero   = zero.to(dtype).to(device)

	if(lp_norm==1):
		shrink_op = lambda x, beta: torch.sign(x)*torch.nn.functional.relu(torch.abs(x) - 1./beta) 
	else:
		shrink_op = lambda x, beta,p=lp_norm: torch.sign(x)*torch.nn.functional.relu(torch.abs(x) - (1./beta)*torch.pow(torch.abs(x), p-1))

	best_error = 1e4
	for i in range(iters):
		W_q   = torch.round(W_f*scale + zero).clamp(min_max[0], min_max[1])
		W_r   = (W_q - zero)/scale
		W_e   = shrink_op(W_f - W_r, beta)
		zero  = torch.mean(W_q - (W_f - W_e)*scale, axis=axis, keepdim=True)
		beta *= kappa

		current_error = float(torch.abs(W_f - W_r).mean())
		if(verbose): 
			print(i, np.round(current_error, 6))
		if(current_error < best_error):
			best_error = current_error
		else:
			break

	scale = scale.to(tensor.device)
	zero  = zero.to(tensor.device)
	del W_f, W_q, W_r, W_e
	torch.cuda.empty_cache()

	return scale, zero 

optimize_weights_proximal = optimize_weights_proximal_legacy
#optimize_weights_proximal = optimize_weights_proximal_v2

#SGD solver  || W - dequantize(quantize(W))||_1 (p=1 only)
def optimize_weights_autograd(tensor, scale, zero, min_max, axis=0, device='cuda', opt_params={'lr':2e-3, 'iters':2500}, verbose=False): 
	W_f             = tensor.to(device) 
	params          = {}
	params['scale'] = torch.nn.Parameter(scale.float().to(device), requires_grad=True)
	params['zero']  = torch.nn.Parameter(zero.float().to(device),  requires_grad=True)
	optimizer       = torch.optim.AdamW([params[k] for k in params], lr=opt_params['lr'], betas=(0.9, 0.99), eps=1e-06, weight_decay=0.)

	def _loss_fct(output, target):
		return torch.mean(torch.abs(target - output)) #L1

	def _fake_quant():
		#Quantize
		W_q = torch.round(W_f*params['scale'] + params['zero']).clamp(min_max[0], min_max[1])
		#Dequantize
		W_r = (W_q - params['zero'])/params['scale'] 
		return W_r

	with torch.no_grad():
		_init_loss = _loss_fct(_fake_quant(), W_f).item()

	def _step():
		optimizer.zero_grad()
		loss    = _loss_fct(_fake_quant(), W_f)
		loss.backward()
		optimizer.step()
		return  np.round(loss.item(), 10)

	for i in range(opt_params['iters']):
		l = _step()
		if(verbose and (i%100)==0): print(i, l)
		
	with torch.no_grad():
		_final_loss = _loss_fct(_fake_quant(), W_f).item()

	if(_final_loss<_init_loss):
		for k in params: params[k] = params[k].data.detach().to(tensor.device)
	else:
		if(verbose): print('optimization failed...')
		params = {'scale':scale, 'zero':zero}

	del W_f	
	torch.cuda.empty_cache()
	return params['scale'], params['zero']



import torch 
import numpy as np 
from .quantize import HQQLinear, HQQBackend, Quantizer

def _get_dense_param(in_features, out_features, device='cuda', trainable=True, dtype=torch.bfloat16):
	W = torch.nn.Linear(in_features, out_features, bias=None).weight.data.t().to(dtype).to(device).contiguous()
	return torch.nn.Parameter(W, requires_grad=trainable)

class HQQLinearLoRA(torch.nn.Module):
	def __init__(self, linear_layer, peft_config):
		super().__init__()

		#Device
		self.device        = linear_layer.device if hasattr(linear_layer, 'device') else next(linear_layer.parameters()).device
		self.train_dtype   = peft_config['train_dtype'] if ('train_dtype' in peft_config) else torch.float

		#Linear layer
		self.linear_layer  = linear_layer
		self.in_features   = linear_layer.in_features
		self.out_features  = linear_layer.out_features
		
		#Bias
		self.bias = None if (linear_layer.bias is None) else linear_layer.bias.clone()
		self.linear_layer.bias = None
		peft_config['train_bias'] = peft_config['train_bias'] if ('train_bias' in peft_config) else False
		if(self.bias is not None):
			self.bias = torch.nn.Parameter(self.bias, requires_grad=peft_config['train_bias'])
		if((self.bias is None) and peft_config['train_bias']):
			self.bias = torch.nn.Parameter(torch.zeros((self.out_features,), device=self.device, dtype=self.train_dtype), requires_grad=True)

		#Dropout
		if('dropout' in peft_config):
			self.peft_drop  = torch.nn.Dropout(p=peft_config['dropout']) if (peft_config['dropout']>0.) else torch.nn.Identity()
		else:
			self.peft_drop  = torch.nn.Identity()

		#LoRA A/B
		self.peft_config = peft_config
		self.lora_alpha  = peft_config['lora_alpha']
		self.r           = peft_config['r']
		self.scaling     = self.lora_alpha/self.r
		self.lora_A      = _get_dense_param(self.in_features, self.r,  device=self.device, trainable=True, dtype=self.train_dtype) 
		self.lora_B      = _get_dense_param(self.r, self.out_features, device=self.device, trainable=True, dtype=self.train_dtype) 

		#LoRA weights init
		if('lora_init' in peft_config):
			#Set lora init
			assert (peft_config['lora_init']['lora_A'].shape[0], peft_config['lora_init']['lora_B'].shape[1])==(self.in_features, self.out_features), \
					"Invalid init LoRA weight shapes. Expected: lora_A: " + str(self.in_features) + " x r , lora_B: r x " + str(self.out_features)  + ")"
			self.lora_A.data = peft_config['lora_init']['lora_A'].to(self.train_dtype).to(self.device)
			self.lora_B.data = peft_config['lora_init']['lora_B'].to(self.train_dtype).to(self.device)
		else:
			#Init weights, as as the original LoRA implementation 
			torch.nn.init.kaiming_uniform_(self.lora_A, a=np.sqrt(5))
			torch.nn.init.zeros_(self.lora_B)

	def forward(self, x):
		x_dtype = x.dtype 

		#Forward with base linear 
		out = self.linear_layer(x)
		
		#LoRA
		out += (torch.matmul(torch.matmul(self.peft_drop(x.to(self.lora_A.dtype)), self.lora_A), self.lora_B)*self.scaling).to(x_dtype)

		#Bias
		if(self.bias is not None):
			out += self.bias

		out = out.to(x_dtype)
		
		return out

	def merge_and_quantize(self, quant_config):

		#Get initial weights
		W = self.linear_layer(torch.eye(self.in_features, device=self.device, dtype=torch.float16)).t() #== self.linear_layer.dequantize()

		#Merge weights
		W += (torch.matmul(self.lora_A.data, self.lora_B.data)*self.scaling).t().to(W.dtype)

		#New HQQ layer
		new_hqq_layer = HQQLinear(None, quant_config)
		new_hqq_layer.bias = None if (self.bias is None) else self.bias.clone() 
		new_hqq_layer.quantize(W, **quant_config)

		return new_hqq_layer

	def cast(self, dtype=torch.float16):
		self.lora_A.data = self.lora_A.data.to(dtype)
		self.lora_B.data = self.lora_B.data.to(dtype)
		if(self.bias is not None):
			self.bias.data = self.bias.data.to(dtype)
		return self

	def state_dict(self):
		return {'lora_A':self.lora_A.data, 'lora_B':self.lora_B.data, 'scaling':self.scaling, 'bias':self.bias}

	def load_state_dict(self, state_dict):
		self.lora_A.data = state_dict['lora_A'].data.to(self.device)
		self.lora_B.data = state_dict['lora_B'].data.to(self.device)
		self.scaling     = state_dict['scaling']
		if(state_dict['bias'] is not None):
			self.bias.data = state_dict['bias'].data.to(self.device)


#LoRA with fake quantization 
class HQQLinearLoRAWithFakeQuant(HQQLinearLoRA):
	def __init__(self, linear_layer, peft_config):
		super(HQQLinearLoRAWithFakeQuant, self).__init__(linear_layer, peft_config)
		self.quant_param = peft_config['quant_param']

	#@torch.no_grad()
	#@torch.compile()
	def fake_quant(self, weight):
		if(self.quant_param):
			W_q, meta  = Quantizer.quantize(weight, **self.quant_param, bitpack=False) #todo: clone() tensor
			weight_est = Quantizer.dequantize(W_q, meta)
		else:
			weight_est = weight
		return weight_est

	def forward(self, x):
		x_dtype = x.dtype
	
		#Get initial weights
		W = self.linear_layer(torch.eye(self.in_features, device=self.device, dtype=x_dtype)).t() #== self.linear_layer.dequantize()

		#Merge weights
		W += (torch.matmul(self.lora_A, self.lora_B)*self.scaling).t().to(W.dtype)

		#Fake quant
		W = self.fake_quant(W).to(x_dtype)

		#Matmul
		out = torch.matmul(x, W.t())

		#Bias
		if(self.bias is not None):
			out += self.bias

		out = out.to(x_dtype)

		return out 

#Experimental
class HQQLinearGroupedProj(torch.nn.Module):
	def __init__(self, linear_layer, peft_config):
		super().__init__()

		#Device
		self.device        = linear_layer.device if hasattr(linear_layer, 'device') else next(linear_layer.parameters()).device
		self.train_dtype   = peft_config['train_dtype'] if ('train_dtype' in peft_config) else torch.float

		#Linear layer
		self.linear_layer  = linear_layer
		self.in_features   = linear_layer.in_features
		self.out_features  = linear_layer.out_features
		self.bias          = None if (linear_layer.bias is None) else linear_layer.bias.clone()

		#Turn-off bias in the linear layer
		self.linear_layer.bias = None

		#Group proj
		self.peft_config = peft_config
		self.proj_size   = peft_config['proj_size']
		self.proj_num    = peft_config['proj_num']
		self.proj        = torch.nn.Parameter(torch.stack([torch.eye(self.proj_size, dtype=self.train_dtype, device=self.device)]*self.proj_num))
		if(peft_config['zero_trainable']):
			self.linear_layer.meta['zero'] = torch.nn.Parameter(self.linear_layer.meta['zero'].to(self.train_dtype), requires_grad=True)

	@torch.compile()
	def forward(self, x):
		x_dtype = x.dtype 

		#Forward with base linear 
		with torch.no_grad():
			W = self.linear_layer.dequantize().clone()
		#W = self.linear_layer(torch.eye(self.in_features, device=self.device, dtype=x_dtype)).t()
		shape  = W.shape

		#Grouped proj
		proj_b, gs = self.proj.shape[0], self.proj.shape[1]
		W          = torch.matmul(self.proj, W.reshape((proj_b, gs, -1)).to(self.proj.dtype)).to(x_dtype).reshape(shape)  

		#Matmul
		out = torch.matmul(x, W.t())

		#Bias
		if(self.bias is not None):
			out += self.bias
		
		out = out.to(x_dtype)

		return out

	def cast(self, dtype=torch.float16):
		self.proj.data = self.proj.data.to(dtype)
		self.linear_layer.meta['zero'] = self.linear_layer.meta['zero'].to(dtype)
		if(self.bias is not None):
			if(self.bias.requires_grad):
				self.bias.data = self.bias.data.to(dtype)
			else:
				self.bias = self.bias.to(dtype)
		return self

	def state_dict(self):
		return {'proj':self.proj.data, 'bias':self.bias, 'peft_config':self.peft_config}

	def load_state_dict(self, state_dict):
		self.proj.data   = state_dict['proj'].data.to(self.device)
		self.bias        = state_dict['bias'] if ('bias' in state_dict) else None
		self.bias        = self.bias.to(self.device) if (self.bias is not None) else None
		self.peft_config = state_dict['peft_config']


_HQQ_LORA_CLASSES = [HQQLinearLoRA, HQQLinearLoRAWithFakeQuant, HQQLinearGroupedProj]
_HQQ_LORA_MAPPING = {'default':HQQLinearLoRA, 'lora_with_fakequant':HQQLinearLoRAWithFakeQuant, 'grouped_proj':HQQLinearGroupedProj}

def is_hqq_lora_layer(layer):
	return type(layer) in _HQQ_LORA_CLASSES

##################################################################################################################
def autoname_modules(model):
	for name, module in model.named_modules():
		module.name = name

#Patching functions
def patch_linear_add_peft(layer, patch_params):
	_peft_config = patch_params
	if(_peft_config):
		lora_type = _peft_config['lora_type'] if ('lora_type' in _peft_config) else 'default'
		new_layer = _HQQ_LORA_MAPPING[lora_type](layer, _peft_config)
	else:
		new_layer = layer
	return new_layer

def patch_linear_merge_peft(layer, patch_params):
	_quant_config = patch_params
	if(_quant_config):
		new_layer = layer.merge_and_quantize(_quant_config)
		del layer
		cleanup()
	else:
		new_layer = layer
	return new_layer

def patch_linear_cast_peft(layer, patch_params):
	if(is_hqq_lora_layer(layer)):
		layer.cast(patch_params)
	return layer

#Putting it all together
class PeftUtils:

	@classmethod
	def get_base_class(cls, model, base_class):
		#Get base class
		if((base_class is None) and hasattr(model, 'base_class')):
			base_class = model.base_class

		assert (base_class is not None), "You need to provide the base HQQ class (LlamaHQQ, MixtralHQQ, etc.) as model.base_class or as an argument base_class=LlamaHQQ"
		return base_class
	
	@classmethod
	def add_lora(cls, model, lora_params, base_class=None, verbose=True):

		#Base classs
		base_class = cls.get_base_class(model, base_class)

		#Freeze
		for param in model.parameters():
			param.requires_grad = False

		#Patch
		base_class.patch_linearlayers(model, patch_linear_add_peft, lora_params, verbose=verbose)

		#Rename modules
		autoname_modules(model)

		#Default backprop backend
		HQQLinear.set_backend(HQQBackend.PYTORCH_BACKPROP)

	@classmethod
	def merge_lora(cls, model, merge_lora_params, base_class=None, verbose=True):
		#Base classs
		base_class = cls.get_base_class(model, base_class)

		#Patch
		base_class.patch_linearlayers(model, patch_linear_merge_peft, merge_lora_params, verbose=verbose)

	@classmethod
	def cast_lora_weights(cls, model, dtype, base_class=None, verbose=True):
		#Base classs
		base_class = cls.get_base_class(model, base_class)

		#Linear tags
		linear_tags = base_class.get_linear_tags()

		#Patch
		base_class.patch_linearlayers(model, 
									  patch_linear_cast_peft, 
									  dict([(linear_tag, dtype) for linear_tag in linear_tags]), 
									  verbose=verbose)


	@classmethod
	def save_lora_weights(cls, model, filename, base_class=None, verbose=True):
		#Base classs
		base_class = cls.get_base_class(model, base_class)

		lora_global_params = {}
		def _patch_linear_save_weights(layer, patch_params, return_layer=True):
			if(is_hqq_lora_layer(layer)):
				lora_global_params[layer.name] = layer.state_dict()
			if(return_layer): return layer

		#Linear tags
		linear_tags = base_class.get_linear_tags()

		#Patch
		base_class.patch_linearlayers(model, 
									  _patch_linear_save_weights, 
									  dict([(linear_tag, None) for linear_tag in linear_tags]), 
									  verbose=verbose)

		#save
		torch.save(lora_global_params, filename)

	@classmethod
	def load_lora_weights(cls, model, filename, base_class=None, verbose=True):
		#Base classs
		base_class = cls.get_base_class(model, base_class)

		lora_global_params = torch.load(filename, map_location='cpu')

		def _patch_linear_load_weights(layer, patch_params, return_layer=True):
			if(is_hqq_lora_layer(layer)):
				layer.load_state_dict(lora_global_params[layer.name])
			if(return_layer): return layer

		#Linear tags
		linear_tags = base_class.get_linear_tags()

		#Patch
		base_class.patch_linearlayers(model, 
									  _patch_linear_load_weights, 
									  dict([(linear_tag, None) for linear_tag in linear_tags]), 
									  verbose=verbose)
